#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 11:05:24 2017

@author: Yoel
"""
import numpy as np
import biosteam as bst
from biosteam import units
from biosteam import main_flowsheet as F
from ._process_settings import price

__all__ = (
    'create_juicing_system',
    'create_ethanol_production_system',
    'create_ethanol_separation_units',
    'create_system',
)

# %% Juicing and evaporation

def create_juicing_system(ID='juicing_sys', sugar_solution_ID=''):
    ### Streams ###
    chemicals = bst.settings.get_chemicals()
    
    z_mass_sugarcane = chemicals.kwarray(
        dict(Glucose=0.0120811,
             Lignin=0.0327653,
             Solids=0.015,
             Sucrose=0.136919,
             Ash=0.006,
             Cellulose=0.0611531,
             Hemicellulose=0.036082,
             Water=0.7)
    )
    
    sugarcane = bst.Stream('sugarcane',
                           flow=333334 * z_mass_sugarcane,
                           units='kg/hr',
                           price=price['Sugar cane'])
    
    enzyme = bst.Stream('enzyme',
                        Cellulose=100, Water=900, units='kg/hr',
                        price=price['Protease'])
    
    imbibition_water = bst.Stream('imbibition_water',
                                  Water=87023.35, units='kg/hr',
                                  T = 338.15)
    
    H3PO4 = bst.Stream('H3PO4',
                       H3PO4=74.23, Water=13.10, units='kg/hr',
                       price=price['H3PO4'])  # to T203
    
    lime = bst.Stream('lime',
                      CaO=333.00, Water=2200.00, units='kg/hr',
                      price=price['Lime'])  # to P5
    
    polymer = bst.Stream('polymer',
                         Flocculant=0.83, units='kg/hr',
                         price=price['Polymer'])  # to T205
    
    rvf_wash_water = bst.Stream('rvf_wash_water',
                                Water=16770, units='kg/hr',
                                T=363.15)  # to C202
    
    ### Unit operations ###
    
    # Feed the shredder
    U101 = units.ConveyingBelt('U101', ins=sugarcane)
    
    # Separate metals
    U102 = units.MagneticSeparator('U102', ins=U101-0)
    
    # Shredded cane
    U103 = units.Shredder('U103', ins=U102-0)
    
    # Hydrolyze starch
    T201 = units.EnzymeTreatment('T201', T=323.15)  # T=50
    
    # Finely crush lipid cane
    U201 = units.CrushingMill('U201',
                              split=dict(Ash=0.92,
                                         Cellulose=0.92,
                                         Glucose=0.04,
                                         Hemicellulose=0.92,
                                         Lignin=0.92,
                                         Sucrose=0.04,
                                         Solids=1),
                              moisture_content=0.5)
    
    # Convey out bagasse
    U202 = units.ConveyingBelt('U202', ins=U201.outs[0], outs='Bagasse')
    
    # Mix in water
    M201 = units.Mixer('M201')
    # crushing_mill_recycle_sys = bst.System('crushing_mill_recycle_sys',
    #                                path=(U201, S201, M201),
    #                                recycle=M201-0)
    
    # Screen out fibers
    S201 = units.VibratingScreen('S201',
                                 split=dict(Ash=0.35,
                                            Cellulose=0.35,
                                            Glucose=0.88,
                                            Hemicellulose=0.35,
                                            Lignin=0.35,
                                            Solids=0,
                                            Sucrose=0.88,
                                            Water=0.88))
    
    # Store juice before treatment
    T202 = units.StorageTank('T202', tau=4, vessel_material='Carbon steel')
    
    # Heat up before adding acid
    H201 = units.HXutility('H201', T=343.15)
    
    # Mix in acid
    T203 = units.MixTank('T203')
    
    # Pump acid solution
    P201 = units.Pump('P201')
    
    # Mix lime solution
    T204 = units.MixTank('T204', tau=0.10)
    P202 = units.Pump('P202')
    
    # Blend acid lipid solution with lime
    T205 = units.MixTank('T205', tau=0.10)
    
    # Mix recycle
    M202 = units.Mixer('M202')
    
    # Heat before adding flocculant
    H202 = units.HXutility('H202', T=372.15)
    
    # Mix in flocculant
    T206 = units.MixTank('T206')
    T206.tau = 0.10
    
    # Separate residual solids
    C201 = units.Clarifier('C201',
                           split=dict(Ash=0,
                                      CaO=0,
                                      Cellulose=0,
                                      Flocculant=0.522,
                                      Glucose=0.522,
                                      Hemicellulose=0,
                                      Lignin=0,
                                      H3PO4=0.522,
                                      Sucrose=0.522,
                                      Water=0.522))
    
    # Remove solids as filter cake
    C202 = units.RVF('C202', 
                     outs=('filter_cake', ''),
                     moisture_content=0.80,
                     split=dict(Ash=0.85,
                                CaO=0.85,
                                Cellulose=0.85,
                                Glucose=0.01,
                                Hemicellulose=0.85,
                                Lignin=0.85,
                                Sucrose=0.01))
    P203 = units.Pump('P203')
    
    
    # Screen out small fibers from sugar stream
    S202 = units.VibratingScreen('S202', outs=(sugar_solution_ID, 'fiber_fines'),
                                 split=dict(Ash=1.0,
                                            CaO=1.0,
                                            Cellulose=1.0,
                                            Flocculant=0.0,
                                            Glucose=0.998,
                                            Hemicellulose=1.0,
                                            Lignin=1.0,
                                            H3PO4=1.0,
                                            Sucrose=0.998,
                                            Water=0.998))
    S202.mesh_opening = 2
    
    ### Process specifications ###
    
    # Specifications dependent on lipid cane flow rate
    def correct_flows():
        U103._run()
        F_mass = sugarcane.F_mass
        # correct enzyme, lime, phosphoric acid, and imbibition water
        enzyme.imass['Cellulose', 'Water'] = 0.003 * F_mass * np.array([0.1, 0.9])
        lime.imass['CaO', 'Water'] = 0.001 * F_mass * np.array([0.046, 0.954])
        H3PO4.imass['H3PO4', 'Water'] = 0.00025 * F_mass
        imbibition_water.imass['Water'] = 0.25* F_mass
    
    U103.specification = correct_flows
    
    # Specifications within a system
    def correct_wash_water():
        P202._run()
        solids = P202.outs[0].imol['Ash', 'CaO', 'Cellulose',
                                   'Hemicellulose', 'Lignin'].sum()
        rvf_wash_water.imol['Water'] = 0.0574 * solids
    
    P202.specification = correct_wash_water
    
    ### System set-up ###
    
    (U103-0, enzyme)-T201
    (T201-0, M201-0)-U201-1-S201-0-T202
    (S201-1, imbibition_water)-M201
    
    T202-0-H201
    (H201-0, H3PO4)-T203-P201
    (P201-0, lime-T204-0)-T205-P202
    (P202-0, P203-0)-M202-H202
    (H202-0, polymer)-T206-C201
    (C201-1, rvf_wash_water)-C202-1-P203
    C201-0-S202
    
    return bst.System(ID,
        [U101,
         U102,
         U103,
         T201,
         bst.System("juice_extraction_sys",
            [U201,
             S201,
             M201],
            recycle=M201-0),
         T202,
         H201,
         T203,
         P201,
         T204,
         T205,
         P202,
         bst.System('juice_separation_sys',
            [M202,
             H202,
             T206,
             C201,
             C202,
             P203],
            recycle=P203-0),
         S202,
         U202],
    )



# %% Ethanol separation

def create_ethanol_separation_units(degassed_beer=None,
                                    ethanol_product=None,
                                    denaturant=None,
                                    IDs={}):
    if not degassed_beer: # Feedstock
        degassed_beer = bst.Stream(
            ID='degassed_beer', 
            phase='l', T=348.34, P=101325, 
            Water=9.65e+04, Ethanol=2.255e+04, Glucose=4916, 
            H3PO4=83.33, Yeast=103, units='kg/hr'
        )
    if not ethanol_product:
        ethanol_product = bst.Stream('ethanol')
    if not denaturant:
        denaturant = bst.Stream('denaturant')
    
    # Beer column
    x_bot = 3.910570816782338e-06
    y_top = 0.2811210085806504
    D302 = units.BinaryDistillation(IDs.get('Beer column', 'D302'), P=101325,
                                y_top=y_top, x_bot=x_bot, k=1.25, Rmin=0.02,
                                LHK=('Ethanol', 'Water'))
    D302.tray_material = 'Stainless steel 304'
    D302.vessel_material = 'Stainless steel 304'
    D302.boiler.U = 1.85
    P302 = units.Pump(IDs.get('Beer column bottoms product pump', 'P302'))
    
    # Mix ethanol Recycle (Set-up)
    M303 = units.Mixer(IDs.get('Recycle mixer', 'M303'))
    
    y_top = 0.8080462715940735
    D303 = units.BinaryDistillation(IDs.get('Distillation', 'D303'), P=101325,
                                y_top=y_top, x_bot=x_bot, k=1.25,
                                LHK=('Ethanol', 'Water'),
                                tray_material='Stainless steel 304',
                                vessel_material='Stainless steel 304',
                                is_divided=True)
    D303.boiler.U = 1.85
    P303 = units.Pump(IDs.get('Distillation bottoms product pump', 'P303'))
    
    # Superheat vapor for mol sieve
    H303 = units.HXutility(IDs.get('Heat exchanger to superheat vapor to molecular sieves', 'H303'),
                           T=115+273.15, V=1)
    
    # Molecular sieve
    U301 = units.MolecularSieve(IDs.get('Molecular sieves', 'U301'),
                                split=(2165.14/13356.04, 1280.06/1383.85),
                                order=('Ethanol', 'Water'))
    
    # Condense ethanol product
    H304 = units.HXutility(IDs.get('Ethanol condenser', 'H304'), 'S149', V=0, T=340.)
    T302 = units.StorageTank(IDs.get('Ethanol day tank', 'T302'), tau=12,
                             vessel_type='Floating roof',
                             vessel_material='Carbon steel')
    P304 = units.Pump(IDs.get('Ethanol day tank pump', 'P304'))
    
    # Storage for gasoline
    T303 = units.StorageTank(IDs.get('Denaturant storage', 'T303'), tau=7*24,
                             vessel_type='Floating roof',
                             vessel_material='Carbon steel')
    P305 = units.Pump(IDs.get('Denaturant pump', 'P305'))
    
    # Denatured ethanol product
    M304 = units.Mixer(IDs.get('Ethanol-denaturant mixer', 'M304'))
    T304 = units.StorageTank(IDs.get('Product tank', 'T304'),
                             vessel_type='Floating roof',
                             vessel_material='Carbon steel',
                             tau=6.5*24, outs=ethanol_product)
    
    ### Ethanol system set-up ###
    
    degassed_beer-D302-1-P302
    (D302-0, U301-0)-M303-0-D303-0-H303-U301
    D303-1-P303
    
    def adjust_denaturant():
        P304._run()
        pure_ethanol = P304.outs[0]
        denaturant.imol['Octane'] = 0.021*pure_ethanol.F_mass/114.232
    
    P304.specification = adjust_denaturant
    U301-1-H304-0-T302-0-P304
    denaturant-T303-P305
    (P305-0, P304-0)-M304-T304

    return {
        'Beer column': D302,
        'Beer column bottoms product pump': P302,
        'Ethanol dehydration system': 
            bst.System('ethanol_dehydration_sys',
                [M303,
                 D303,
                 H303,
                 U301],
                 recycle=U301-0
            ),
        'Recycle mixer': M303,
        'Distillation': D303,
        'Heat exchanger to superheat vapor to molecular sieves': H303,
        'Molecular sieves': U301,
        'Ethanol condenser': H304,
        'Ethanol day tank': T302, 
        'Ethanol day tank pump': P304,
        'Denaturant storage': T303, 
        'Denaturant pump': P305, 
        'Ethanol-denaturant mixer': M304,
        'Product tank': T304,
        'Distillation bottoms product pump': P303,
    }
        

# %% Ethanol production section (fermentation and separations)

def create_ethanol_production_system(ID='ethanol_production_sys',
                                     sugar_solution=None):
    ### Streams ###
    
    # Fresh water
    stripping_water = bst.Stream('stripping_water', Water=5000, units='kg/hr')
    
    # Gasoline
    denaturant = bst.Stream('denaturant', Octane=230.69,
                            units='kg/hr', price=price['Gasoline'])
    
    if not sugar_solution: # Feedstock
        sugar_solution = bst.Stream('sugar_solution',
            Glucose = 3802,
            Sucrose = 4.309e+04,
            Water   = 2.59e+05,
            H3PO4   = 83.33,
            units = 'kg/hr',
            T = 372,
        )
    
    # Yeast
    yeast = bst.Stream('yeast', Water=24700, DryYeast=10300, units='kg/hr')
    
    # Ethanol product
    ethanol = bst.Stream('ethanol', price=price['Ethanol'])
    
    ### Units ###
    
    # Split sugar solution
    S301 = units.Splitter('S301',
                        split=0.265)
    
    # Concentrate sugars
    F301 = units.MultiEffectEvaporator('F301',
                                       P=(101325, 73581, 50892, 32777),
                                       V=0.95) # fraction evaporated
    F301.components['condenser'].U = 1.85
    # Note: value of steam ~ 6.86 for the following 
    # (101325, 73580.467, 50891.17, 32777.406, 19999.925, 11331.5),
    
    # Mix sugar solutions
    M301 = units.Mixer('M301')
    
    # Cool for fermentation
    H301 = units.HXutility('H301', T=295.15)
    
    # Ethanol Production
    R301 = units.Fermentation('R301', outs=('CO2', ''), tau=9, efficiency=0.90, N=4) 
    T301 = units.StorageTank('T301', tau=4, vessel_material='Carbon steel')
    T301.line = 'Beer tank'
    
    D301 = units.VentScrubber('D301', ins=(stripping_water, R301-0), 
                              outs=('vent', ''),
                              gas=('CO2',))
    
    # Separate 99% of yeast
    C301 = units.SolidsCentrifuge('C301', outs=('', 'recycle_yeast'),
                                split=(1, 0.99999, 1, 0.96, 0.01),
                                order=('Ethanol', 'Glucose', 'H3PO4', 
                                       'Water', 'DryYeast'),
                                solids=('DryYeast',))
    
    # Mix in Water
    M302 = units.Mixer('M302')
    P301 = units.Pump('P301')
    
    # Heat up before beer column
    # Exchange heat with stillage
    H302 = units.HXprocess('H302', phase0='l', phase1='l', U=1.28)
    
    dct = create_ethanol_separation_units(
        degassed_beer=H302-0,
        ethanol_product=ethanol,
        denaturant=denaturant)
    D302 = dct['Beer column']
    P302 = dct['Beer column bottoms product pump']
    ethanol_dehydration_sys = dct['Ethanol dehydration system']
    H304 = dct['Ethanol condenser']
    T302 = dct['Ethanol day tank']
    P304 = dct['Ethanol day tank pump']
    T303 = dct['Denaturant storage']
    P305 = dct['Denaturant pump']
    M304 = dct['Ethanol-denaturant mixer']
    T304 = dct['Product tank']
    P303 = dct['Distillation bottoms product pump']
    
    # Waste water
    M305 = units.Mixer('M305', outs='wastewater')
    
    # Yeast mixing
    T305 = units.MixTank('T305')
    T305.tau = 0.1
    yeast-T305
    
    # Multi-effect evaporator pumps
    P306 = units.Pump('P306')
    
    
    ### Ethanol system set-up ###
    
    sugar_solution-S301-1-F301-0-P306
    (S301-0, P306-0)-M301-H301
    (H301-0, yeast-T305-0)-R301-1-T301-0-C301
    (C301-0, D301-1)-M302-P301
    (P303-0, F301-1, (P301-0, P302-0)-H302-1)-M305
    
    ### System ###
    
    return bst.System(ID, 
                [S301, 
                 F301, 
                 P306, 
                 M301, 
                 H301, 
                 T305,
                 R301,
                 T301, 
                 C301, 
                 M302, 
                 P301,
                 bst.System('beer_column_heat_integration',
                     [H302,
                      D302,
                      P302],
                     recycle=P302-0),
                 ethanol_dehydration_sys,
                 H304,
                 T302, 
                 P304,
                 T303, 
                 P305, 
                 M304,
                 T304, 
                 D301,
                 P303, 
                 M305])


# %% Complete system

def create_system(ID='sugarcane_sys'):
    s = F.stream
    u = F.unit
    
    ### Juicing section ###
    
    juicing_system = create_juicing_system()
    
    ### Ethanol section ###
    
    ethanol_production_sys = create_ethanol_production_system(sugar_solution=u.S202-0)
    
    ### Facilities ###    
    
    BT = units.BoilerTurbogenerator('BT',
                                    (u.U202-0, '', 'boiler_makeup_water', 'natural_gas', '', ''),
                                    boiler_efficiency=0.80,
                                    turbogenerator_efficiency=0.85)
    
    CT = units.CoolingTower('CT')
    makeup_water_streams = (s.cooling_tower_makeup_water,
                            s.boiler_makeup_water)
    process_water_streams = (s.imbibition_water,
                             s.rvf_wash_water,
                             s.stripping_water,
                             *makeup_water_streams)
    makeup_water = bst.Stream('makeup_water', price=0.000254)
    
    CWP = units.ChilledWaterPackage('CWP')
    PWC = units.ProcessWaterCenter('PWC',
                                   (bst.Stream(), makeup_water),
                                   (),
                                   None,
                                   makeup_water_streams,
                                   process_water_streams)
    
    ### System ###
    
    return bst.System(ID,
        [juicing_system,
         ethanol_production_sys],
        facilities=(CWP, BT, CT, PWC),
    )



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .. import PY37
from . import (_process_settings,
               _chemicals,
               _system,
               _tea,
)

__all__ = [*_process_settings.__all__,
           *_chemicals.__all__,
           *_system.__all__,
           *_tea.__all__,
           'sugarcane_sys',
           'sugarcane_tea', 
           'flowsheet',
]

from ._process_settings import *
from ._chemicals import *
from ._system import *
from ._tea import *

_system_loaded = False
_chemicals_loaded = False

def load():
    if not _chemicals_loaded: _load_chemicals()
    try: 
        _load_system()
    finally: 
        dct = globals()
        dct.update(flowsheet.system.__dict__)
        dct.update(flowsheet.stream.__dict__)
        dct.update(flowsheet.unit.__dict__)

def _load_chemicals():
    global chemicals, _chemicals_loaded
    chemicals = create_chemicals()
    _chemicals_loaded = True

def _load_system():
    import biosteam as bst
    from biosteam import main_flowsheet as F
    global sugarcane_sys, sugarcane_tea, specs, flowsheet, _system_loaded
    flowsheet = bst.Flowsheet('sugarcane')
    F.set_flowsheet(flowsheet)
    bst.settings.set_thermo(chemicals)
    load_process_settings()
    sugarcane_sys = create_system()
    sugarcane_sys.simulate()
    sugarcane_tea = create_tea(sugarcane_sys)
    sugarcane_tea.IRR = sugarcane_tea.solve_IRR()
    _system_loaded = True
  
if PY37:
    def __getattr__(name):
        if not _chemicals_loaded:
            _load_chemicals()
            if name == 'chemicals': return chemicals
        if not _system_loaded: 
            try:
                _load_system()
            except Exception as Error:
                dct = globals()
                dct.update(flowsheet.system.__dict__)
                dct.update(flowsheet.stream.__dict__)
                dct.update(flowsheet.unit.__dict__)
                if name in dct: return dct[name]
                raise Error
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")
else: 
    load()



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import numpy as np
import biosteam as bst
import thermosteam as tmo
import flexsolve as flx
import pytest
from biosteam.process_tools import UnitGroup
from importlib import import_module

__all__ = (
    'test_sugarcane',
    'test_lipidcane',
    'test_cornstover',
    'test_LAOs',
    'test_wheatstraw',
    'test_animal_bedding',
    'test_lactic',
    'test_ethanol_adipic',
    'generate_code',
    'print_results',
)

# Block speed-up for consistent testing. Note that speed-up wrapps functions,
# thus preventing code coverage to be analyzed
bst.speed_up = tmo.speed_up = flx.speed_up = lambda: None 

feedstocks_by_module = {
    'lipidcane': 'lipidcane',
    'cornstover': 'cornstover',
    'sugarcane': 'sugarcane',
    'LAOs': 'glucose',
    'wheatstraw': 'wheatstraw',
    'animal_bedding': 'bedding',
    'lactic': 'feedstock',
    'ethanol_adipic': 'feedstock',
}
products_by_module = {
    'lipidcane': 'ethanol',
    'cornstover': 'ethanol',
    'sugarcane': 'ethanol',
    'LAOs': 'octene',
    'wheatstraw': 'ethanol',
    'animal_bedding': 'ethanol',
    'lactic': 'lactic_acid',
    'ethanol_adipic': 'ethanol',
}

def generate_code(module_name, feedstock_name=None, product_name=None):
    if not feedstock_name:
        feedstock_name = feedstocks_by_module[module_name]
    if not product_name:
        product_name = products_by_module[module_name]
    bst.process_tools.default_utilities()
    module = import_module('biorefineries.' + module_name)
    try: module.load()
    except: pass
    feedstock = getattr(module, feedstock_name)
    product = getattr(module, product_name)
    tea_name = f'{module_name}_tea'
    try:
        tea = getattr(module, tea_name)
    except AttributeError:
        try:
            tea_name = f'{feedstock_name}_tea'
            tea = getattr(module, tea_name)
        except AttributeError:
            tea_name = f'{product_name}_tea'
            tea = getattr(module, tea_name)
    units = UnitGroup('Biorefinery', tea.units)
    IRR = tea.IRR
    sales = tea.sales
    material_cost = tea.material_cost
    installed_equipment_cost = tea.installed_equipment_cost
    utility_cost = tea.utility_cost
    heating_duty = units.get_heating_duty()
    cooling_duty = units.get_cooling_duty()
    electricity_consumption = units.get_electricity_consumption()
    electricity_production = units.get_electricity_production()
    print(f"""def test_{module_name}():
    bst.process_tools.default_utilities()
    from biorefineries import {module_name} as module
    try: module.load()
    except: pass
    feedstock = module.{feedstock_name}
    product = module.{product_name}
    tea = module.{tea_name}
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, {IRR}, rtol=1e-2)
    assert np.allclose(feedstock.price, {feedstock.price}, rtol=1e-2)
    assert np.allclose(product.price, {product.price}, rtol=1e-2)
    assert np.allclose(tea.sales, {sales}, rtol=1e-2)
    assert np.allclose(tea.material_cost, {material_cost}, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, {installed_equipment_cost}, rtol=1e-2)
    assert np.allclose(tea.utility_cost, {utility_cost}, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), {heating_duty}, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), {cooling_duty}, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), {electricity_consumption}, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), {electricity_production}, rtol=1e-2)
    """
    )

def print_results(tea):
    units = UnitGroup('Biorefinery', tea.units)
    print('Sales:', tea.sales)
    print('Material cost:', tea.material_cost)
    print('Installed equipment cost:', tea.installed_equipment_cost)
    print('Utility cost:', tea.utility_cost)
    print('Heating duty:', units.get_heating_duty())
    print('Cooling duty:', units.get_cooling_duty())
    print('Electricity consumption:', units.get_electricity_consumption())
    print('Electricity production:', units.get_electricity_production())
    
def test_sugarcane():
    bst.process_tools.default_utilities()
    from biorefineries import sugarcane as module
    try: module.load()
    except: pass
    feedstock = module.sugarcane
    product = module.ethanol
    tea = module.sugarcane_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.10584901424522822, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.03455, rtol=1e-2)
    assert np.allclose(product.price, 0.789, rtol=1e-2)
    assert np.allclose(tea.sales, 87558012.26683149, rtol=1e-2)
    assert np.allclose(tea.material_cost, 60092629.818772286, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 181877578.28509316, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -9901284.14957534, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 263.5001340760937, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 329.94966728221334, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 10.239976616876824, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 41.974861711669575, rtol=1e-2)

def test_lipidcane():
    bst.process_tools.default_utilities()
    from biorefineries import lipidcane as module
    try: module.load()
    except: pass
    feedstock = module.lipidcane
    product = module.ethanol
    tea = module.lipidcane_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.17931307067069024, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.03455, rtol=1e-2)
    assert np.allclose(product.price, 0.789, rtol=1e-2)
    assert np.allclose(tea.sales, 102369101.03579773, rtol=1e-2)
    assert np.allclose(tea.material_cost, 61760752.25323551, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 224875041.54906383, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -27155273.13884334, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 191.5907142460678, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 294.91214479431926, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 10.217107389763953, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 97.25323924503112, rtol=1e-2)

def test_cornstover():
    bst.process_tools.default_utilities()
    from biorefineries import cornstover as module
    try: module.load()
    except: pass
    feedstock = module.cornstover
    product = module.ethanol
    tea = module.cornstover_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.05158816935126135, rtol=1e-2)
    assert np.allclose(product.price, 0.7329144150557967, rtol=1e-2)
    assert np.allclose(tea.sales, 133072785.0709948, rtol=1e-2)
    assert np.allclose(tea.material_cost, 82902427.63227808, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 220687096.75274014, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -12169917.32871969, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 310.3265668144915, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 358.6403621978023, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 22.773951671296523, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(),48.07370343834068, rtol=1e-2)
    
def test_LAOs():
    bst.process_tools.default_utilities()
    from biorefineries import LAOs as module
    try: module.load()
    except: pass
    feedstock = module.glucose
    product = module.octene
    tea = module.LAOs_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.265, rtol=1e-2)
    assert np.allclose(product.price, 1.3297347335767877, rtol=1e-2)
    assert np.allclose(tea.sales, 165990937.17560306, rtol=1e-2)
    assert np.allclose(tea.material_cost, 135661583.58974993, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 70934744.36669901, rtol=1e-2)
    assert np.allclose(tea.utility_cost, 2808901.449402818, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 40.00054620871385, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 125.37714908516853, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 3.6913734166374796, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 3.69137341663748, rtol=1e-2) 

@pytest.mark.slow
def test_wheatstraw():
    bst.process_tools.default_utilities()
    from biorefineries import wheatstraw as module
    try: module.load()
    except: pass
    feedstock = module.wheatstraw
    product = module.ethanol
    tea = module.wheatstraw_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.056999999999999995, rtol=1e-2)
    assert np.allclose(product.price, 0.9307886482467054, rtol=1e-2)
    assert np.allclose(tea.sales, 129923549.54214548, rtol=1e-2)
    assert np.allclose(tea.material_cost, 63431954.80497392, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 243695705.11266732, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -5195846.43776603, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 214.26465244400836, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 261.68181957021756, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 23.29205112444854, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 34.09357274078234, rtol=1e-2)
    

@pytest.mark.slow
def test_animal_bedding():
    bst.process_tools.default_utilities()
    from biorefineries import animal_bedding as module
    try: module.load()
    except: pass
    feedstock = module.bedding
    product = module.ethanol
    tea = module.bedding_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.003, rtol=1e-2)
    assert np.allclose(product.price, 0.8095823080765768, rtol=1e-2)
    assert np.allclose(tea.sales, 114559124.26970857, rtol=1e-2)
    assert np.allclose(tea.material_cost, 33147673.66201239, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 285644882.33318335, rtol=1e-2)
    assert np.allclose(tea.utility_cost, -1905859.1223573387, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 157.88541329603024, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 198.23737761715043, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 32.31812207544716, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 36.28016728875428, rtol=1e-2)
    
@pytest.mark.slow
def test_lactic():
    bst.process_tools.default_utilities()
    from biorefineries import lactic as module
    try: module.load()
    except: pass
    feedstock = module.feedstock
    product = module.lactic_acid
    tea = module.lactic_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.06287583717512708, rtol=1e-2)
    assert np.allclose(product.price, 1.5290315667995373, rtol=1e-2)
    assert np.allclose(tea.sales, 332346881.2446115, rtol=1e-2)
    assert np.allclose(tea.material_cost, 219074528.09850037, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 321333077.7056334, rtol=1e-2)
    assert np.allclose(tea.utility_cost, 25050313.992286183, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 1829.0904347765866, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 1895.6439958343117, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 45.39087118990755, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 0.0, rtol=1e-2)
    

@pytest.mark.slow
def test_ethanol_adipic():
    bst.process_tools.default_utilities()
    from biorefineries import ethanol_adipic as module
    try: module.load()
    except: pass
    feedstock = module.feedstock
    product = module.ethanol
    tea = module.ethanol_adipic_tea
    units = UnitGroup('Biorefinery', tea.units)
    assert np.allclose(tea.IRR, 0.1, rtol=1e-2)
    assert np.allclose(feedstock.price, 0.06287581915485817, rtol=1e-2)
    assert np.allclose(product.price, 0.8360250946573881, rtol=1e-2)
    assert np.allclose(tea.sales, 219891872.86390597, rtol=1e-2)
    assert np.allclose(tea.material_cost, 120581699.4153729, rtol=1e-2)
    assert np.allclose(tea.installed_equipment_cost, 283387972.8581399, rtol=1e-2)
    assert np.allclose(tea.utility_cost, 14242922.86112985, rtol=1e-2)
    assert np.allclose(units.get_heating_duty(), 296.834875396348, rtol=1e-2)
    assert np.allclose(units.get_cooling_duty(), 247.78343626334382, rtol=1e-2)
    assert np.allclose(units.get_electricity_consumption(), 26.56706595101369, rtol=1e-2)
    assert np.allclose(units.get_electricity_production(), 0.0, rtol=1e-2)
    
### DO NOT DELETE:
### Code commented for legacy purposes
# @pytest.mark.slow
# def test_lactic():
#     from biorefineries import lactic
#     bst.process_tools.default_utilities()
#     system = lactic.system
#     MPSP = system.lactic_acid.price
#     assert np.allclose(MPSP, 1.5285811040727408, atol=0.01)
#     tea = system.lactic_tea
#     assert np.allclose(tea.sales, 332247972.3322271, rtol=0.01)
#     assert np.allclose(tea.material_cost, 218911499.70700422, rtol=0.01)
#     assert np.allclose(tea.installed_equipment_cost, 321744540.84435904, rtol=0.01)
#     assert np.allclose(tea.utility_cost, 24995492.58140952, rtol=0.01)
#     units = UnitGroup('Biorefinery', system.lactic_tea.units)
#     assert np.allclose(system.CHP.system_heating_demand/1e6, 1763.4942302374052, rtol=0.01)
#     assert np.allclose(-system.CT.system_cooling_water_duty/1e6, 1629.0964343101657, rtol=0.01)
#     assert np.allclose(units.get_electricity_consumption(), 45.291535445041546, rtol=0.01)
#     assert np.allclose(units.get_electricity_production(), 0.0)

# @pytest.mark.slow
# def test_ethanol_adipic():
#     bst.process_tools.default_utilities()
#     from biorefineries import ethanol_adipic
#     acid = ethanol_adipic.system_acid
#     MESP = acid.ethanol.price * acid._ethanol_kg_2_gal
#     assert np.allclose(MESP, 2.486866594017598, atol=0.01)
#     tea = acid.ethanol_tea
#     assert np.allclose(tea.sales, 152002785.9542236, rtol=0.01)
#     assert np.allclose(tea.material_cost, 112973059.93909653, rtol=0.01)
#     assert np.allclose(tea.installed_equipment_cost, 202880204.7890376, rtol=0.01)
#     assert np.allclose(tea.utility_cost, -18225095.0315181, rtol=0.01)
#     units = UnitGroup('Biorefinery', acid.ethanol_tea.units)
#     assert np.allclose(acid.CHP.system_heating_demand/1e6, 333.8802418517945, rtol=0.01)
#     assert np.allclose(units.get_cooling_duty(), 327.05990128304114, rtol=0.01)
#     assert np.allclose(units.get_electricity_consumption(), 23.84999102548279, rtol=0.01)
#     assert np.allclose(units.get_electricity_production(), 55.72024685271333, rtol=0.01)
    
#     base = ethanol_adipic.system_base
#     MESP = base.ethanol.price * base._ethanol_kg_2_gal
#     assert np.allclose(MESP, 2.703699730342356, atol=0.01)
#     tea = base.ethanol_adipic_tea
#     assert np.allclose(tea.sales, 219850046.2308626, rtol=0.01)
#     assert np.allclose(tea.material_cost, 120551649.9082640, rtol=0.01)
#     assert np.allclose(tea.installed_equipment_cost, 283345546.0556234, rtol=0.01)
#     assert np.allclose(tea.utility_cost, 14241964.663629433, rtol=0.01)
#     units = UnitGroup('Biorefinery', base.ethanol_adipic_tea.units)
#     assert np.allclose(base.CHP.system_heating_demand/1e6, 296.8322623939439, rtol=0.01)
#     assert np.allclose(units.get_cooling_duty(), 247.79573940245342, rtol=0.01)
#     assert np.allclose(units.get_electricity_consumption(), 26.565278642577344, rtol=0.001)
#     assert np.allclose(units.get_electricity_production(), 0.0)
    
if __name__ == '__main__':
    test_sugarcane()
    test_lipidcane()
    test_cornstover()
    test_wheatstraw()
    test_LAOs()
    test_animal_bedding()
    test_lactic()
    test_ethanol_adipic()







# -*- coding: utf-8 -*-
"""
Example module for calculating ethanol yield

"""
from biorefineries.cornstover.abm import ABM_TEA_model as lignocellulose_model
from biorefineries.corn.abm import ABM_TEA_model as corn_model

lignocellulosic_capacity = 876072883.4242561
corn_capacity = 876072883.4242561
specific_ethanol_volume = 1.267 # L/kg
kg_per_ton = 907.185
def compute_ethanol_yield(plant_capacity, production):
    return (
        specific_ethanol_volume
        * kg_per_ton
        * production
        / plant_capacity
    )


# Yield of Corn
results = corn_model(plant_capacity=corn_capacity)
ethanol_yield = compute_ethanol_yield(corn_capacity, results['Production'])
print('Corn ethanol yield:', round(ethanol_yield), 'L/ton')

# Yield of Corn Stover
results = lignocellulose_model(plant_capacity=lignocellulosic_capacity, cornstover_fraction=1)
ethanol_yield = compute_ethanol_yield(lignocellulosic_capacity, results['Production'])
print('Corn stover ethanol yield:', round(ethanol_yield), 'L/ton')

# Yield of Minscanthus
results = lignocellulose_model(plant_capacity=lignocellulosic_capacity, cornstover_fraction=0)
ethanol_yield = compute_ethanol_yield(lignocellulosic_capacity, results['Production'])
print('Miscanthus ethanol yield:', round(ethanol_yield), 'L/ton')


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .. import PY37
from . import (units,
               utils,
               _chemicals,
               _process_specifications,
               _system,
               _tea,
)

__all__ = [*units.__all__,
           *utils.__all__,
           *_chemicals.__all__,
           *_process_specifications.__all__,
           *_system.__all__,
           *_tea.__all__,
           'LAOs_sys',
           'LAOs_tea', 
           'flowsheet', 
           'unit_groups', 
           'OSBL_unit_group',
           'specs',
           'utils',
]

from .utils import *
from .units import *
from ._chemicals import *
from ._process_specifications import *
from ._system import *
from ._tea import *

_system_loaded = False
_chemicals_loaded = False

def load():
    if not _chemicals_loaded: _load_chemicals()
    _load_system()
    dct = globals()
    dct.update(flowsheet.system.__dict__)
    dct.update(flowsheet.stream.__dict__)
    dct.update(flowsheet.unit.__dict__)

def _load_chemicals():
    global chemicals
    chemicals = create_chemicals()    

def _load_system():
    import biosteam as bst
    from biosteam import main_flowsheet as F
    global LAOs_sys, LAOs_tea, specs, flowsheet, unit_groups, OSBL_unit_group
    global _system_loaded, products
    flowsheet = bst.Flowsheet('LAOs')
    F.set_flowsheet(flowsheet)
    bst.settings.set_thermo(chemicals)
    load_process_settings()
    LAOs_sys = create_system()
    OSBL_units = (F.unit.CWP, F.unit.BT, F.unit.CT, F.unit.T101, F.unit.T102, 
                  F.unit.T103, F.unit.T104, F.unit.T107, F.unit.T108, 
                  F.unit.T109, F.unit.T110)
    LAOs_tea = create_tea(LAOs_sys, OSBL_units)
    for i in LAOs_tea.TEAs: i.duration = (2017, 2047)
    for i in LAOs_tea.TEAs: i.contingency = 0.3
    specs = LAOsProcessSpecifications(LAOs_sys, LAOs_tea)
    UnitGroup = bst.process_tools.UnitGroup
    name_types = {
        'Tanks': bst.Tank,
        'Pumps': bst.Pump,
        'Facilities':bst.Facility,
        'Fermentation': bst.BatchBioreactor,
        'Heat exchangers': bst.HX,
        'Dehydration reactor': units.AdiabaticFixedbedGasReactor,
        'Mixer-settler': (bst.MixerSettler),
        'Centrifuges': (bst.LiquidsCentrifuge,
                        SolidLiquidsSplitCentrifuge),
        'Distillation': bst.BinaryDistillation
    }
    unit_groups = UnitGroup.group_by_types(LAOs_tea.units, name_types)
    OSBL_unit_group = UnitGroup('OSBL', OSBL_units)
    bst.System.molar_tolerance = 0.1
    bst.System.converge_method = 'aitken'
    specs.run_specifications() # Sets process specifications and simulates system
    products = (F('hexene'), F('octene'), F('decene'))
    for i in range(2): set_LAOs_MPSP(get_LAOs_MPSP())
    _system_loaded = True

if PY37:
    def __getattr__(name):
        if not _chemicals_loaded:
            _load_chemicals()
            if name == 'chemicals': return chemicals
        if not _system_loaded: _load_system()
        dct = globals()
        dct.update(flowsheet.system.__dict__)
        dct.update(flowsheet.stream.__dict__)
        dct.update(flowsheet.unit.__dict__)
        if name in dct:
            return dct[name]
        else:
            raise AttributeError(f"module '{__name__}' has no attribute '{name}'")
else:
    load()


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

import os
from doctest import testfile
from biorefineries import (
    sugarcane as sc,
    lipidcane as lc,
    cornstover as cs,
    LAOs as laos,
    lactic,
    ethanol_adipic,
    )

__all__ = (
    'test_sugarcane_readme',
    'test_lipidcane_readme',
    'test_cornstover_readme',
    'test_laos_readme',
    'test_lactic_readme',
    'test_ethanol_adipic_readme',
)

get_readme = lambda module: os.path.join(os.path.dirname(module.__file__), 'README.rst')
test_sugarcane_readme = lambda: testfile(get_readme(sc))
test_lipidcane_readme = lambda: testfile(get_readme(lc))
test_cornstover_readme = lambda: testfile(get_readme(cs))
test_laos_readme = lambda: testfile(get_readme(laos))
test_lactic_readme = lambda: testfile(get_readme(lactic))
test_ethanol_adipic_readme = lambda: testfile(get_readme(ethanol_adipic))

if __name__ == '__main__':
    test_sugarcane_readme()
    test_lipidcane_readme()
    test_cornstover_readme()
    test_laos_readme()
    test_lactic_readme()
    test_ethanol_adipic_readme()



# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 06:42:02 2020

@author: yoelr
"""
from thermosteam import functional as fn
import thermosteam as tmo

__all__ = ('create_chemicals',)

def create_chemicals():
    from biorefineries import sugarcane as sc
    (Water, Ethanol, Glucose, Sucrose, H3PO4, P4O10, CO2, Octane, O2, CH4, 
     Ash, Cellulose, Hemicellulose, Flocculant, Lignin, Solids, DryYeast, CaO) = sc.chemicals
    
    
    
    ### Define common chemicals ###
    Biodiesel = tmo.Chemical('Biodiesel',
                             search_ID='Methyl oleate')
    Methanol = tmo.Chemical('Methanol')
    Glycerol = tmo.Chemical('Glycerol')
    lipidcane_chemicals = sc.chemicals.copy()
    lipidcane_chemicals.extend([Biodiesel, Methanol, Glycerol])
    
    ### Define new chemicals ###
    
    def create_new_chemical(ID, phase='s', **constants):
        solid = tmo.Chemical.blank(ID, phase=phase, **constants)
        lipidcane_chemicals.append(solid)
        return solid
    
    HCl = create_new_chemical('HCl', formula='HCl')
    NaOH = create_new_chemical('NaOH', formula='NaOH')
    NaOCH3 = create_new_chemical('NaOCH3', formula='NaOCH3')
    Lipid = create_new_chemical(
        'Lipid',
        phase = 'l',
        formula = 'C57H104O6',
        Hf = -2193.7e3
    )
    Lipid.Dortmund.set_group_counts_by_name({'CH3':3, 'CH2':41, 'CH':1, 'CH=CH':3, 'CH2COO':3})
    
    ### Fill missing properties ###
    
    # Assume properties are similar for trioleate and tripalmitin
    Tripalmitin = tmo.Chemical('Tripalmitin').at_state(phase='l', copy=True)
    Lipid.copy_models_from(Tripalmitin, ['V', 'sigma', 'kappa', 'Cn'])
    
    # Assume a constant volume for lipid
    lipid_molar_volume = fn.rho_to_V(rho=900, MW=Lipid.MW)
    Lipid.V.add_model(lipid_molar_volume)
    
    # Solubles don't occupy much volume
    for chemical in (HCl, NaOH):
        V = fn.rho_to_V(rho=1e5, MW=chemical.MW)
        chemical.V.add_model(V, top_priority=True)
    
    # Assume sodium methoxide has some of the same properities as methanol
    LiquidMethanol = Methanol.at_state(phase='l', copy=True)
    NaOCH3.copy_models_from(LiquidMethanol, ['V', 'sigma', 'kappa', 'Cn'])
    
    for chemical in lipidcane_chemicals: chemical.default()
    
    lipidcane_chemicals.compile()
    lipidcane_chemicals.set_synonym('Water', 'H2O')
    lipidcane_chemicals.set_synonym('Yeast', 'DryYeast')
    return lipidcane_chemicals

